"""RunLocal API Client Package"""

from .client import RunLocalClient
from .devices import DeviceFilters
from .exceptions import (
    APIError,
    AuthenticationError,
    ConfigurationError,
    DeviceNotAvailableError,
    JobTimeoutError,
    ModelNotFoundError,
    NetworkError,
    RunLocalError,
    TensorError,
    UploadError,
    ValidationError,
)
from .models import (
    BenchmarkData,
    BenchmarkResponse,
    BenchmarkResult,
    BenchmarkStatus,
    Framework,
    Device,
    DeviceUsage,
    JobResult,
    JobType,
    IOType,
    PredictionResponse,
    PredictionResult,
    TensorInfo,
    RuntimeSettings,
)
from .utils import (
    display_benchmark_results,
    display_failed_benchmarks,
    display_model,
)

__all__ = [
    "RunLocalClient",
    "Device",
    "DeviceUsage",
    "DeviceFilters",
    "JobType",
    "JobResult",
    "IOType",
    "TensorInfo",
    "BenchmarkData",
    "BenchmarkResult",
    "BenchmarkResponse",
    "BenchmarkStatus",
    "PredictionResult",
    "PredictionResponse",
    "RuntimeSettings",
    "Framework",
    "RunLocalError",
    "AuthenticationError",
    "APIError",
    "ModelNotFoundError",
    "DeviceNotAvailableError",
    "JobTimeoutError",
    "TensorError",
    "UploadError",
    "ValidationError",
    "NetworkError",
    "ConfigurationError",
    "display_benchmark_results",
    "display_failed_benchmarks",
    "display_model",
]



from enum import Enum
from typing import Optional, Union

from pydantic import BaseModel

from .common import NumThreads


class InferencePreference(str, Enum):
    FAST_SINGLE_ANSWER = "FAST_SINGLE_ANSWER"
    SUSTAINED_SPEED = "SUSTAINED_SPEED"
    BALANCED = "BALANCED"


class InferencePriority(str, Enum):
    AUTO = "AUTO"
    MAX_PRECISION = "MAX_PRECISION"
    MIN_LATENCY = "MIN_LATENCY"
    MIN_MEMORY_USAGE = "MIN_MEMORY_USAGE"


class PerformanceMode(str, Enum):
    DEFAULT = "Default"
    SUSTAINED_HIGH_PERFORMANCE = "SustainedHighPerformance"
    BURST = "Burst"
    HIGH_PERFORMANCE = "HighPerformance"
    POWER_SAVER = "PowerSaver"
    LOW_POWER_SAVER = "LowPowerSaver"
    HIGH_POWER_SAVER = "HighPowerSaver"
    LOW_BALANCED = "LowBalanced"
    BALANCED = "Balanced"
    EXTREME_POWER_SAVER = "ExtremePowerSaver"


class Precision(str, Enum):
    QUANTIZED = "Quantized"
    FP16 = "Fp16"


class PdSession(str, Enum):
    UNSIGNED_PD = "UnsignedPd"
    SIGNED_PD = "SignedPd"


class OptimizationStrategy(str, Enum):
    OPTIMIZE_FOR_INFERENCE = "OptimizeForInference"
    OPTIMIZE_FOR_PREPARE = "OptimizeForPrepare"
    OPTIMIZE_FOR_INFERENCE_O3 = "OptimizeForInferenceO3"


class XNNPACKSettings(BaseModel):
    num_threads: Optional[Union[int, NumThreads]] = None
    QS8: Optional[bool] = None
    QU8: Optional[bool] = None
    FORCE_FP16: Optional[bool] = None
    DYNAMIC_FULLY_CONNECTED: Optional[bool] = None
    VARIABLE_OPERATORS: Optional[bool] = None
    TRANSIENT_INDIRECTION_BUFFER: Optional[bool] = None
    ENABLE_LATEST_OPERATORS: Optional[bool] = None


class GPUSettings(BaseModel):
    is_precision_loss_allowed: Optional[bool] = None
    inference_preference: Optional[InferencePreference] = None
    inference_priority1: Optional[InferencePriority] = None
    inference_priority2: Optional[InferencePriority] = None
    inference_priority3: Optional[InferencePriority] = None
    enable_quant: Optional[bool] = None
    enable_serialization: Optional[bool] = None
    max_delegated_partitions: Optional[int] = None


class QNNSettings(BaseModel):
    performance_mode: Optional[PerformanceMode] = None
    precision: Optional[Precision] = None
    pd_session: Optional[PdSession] = None
    optimization_strategy: Optional[OptimizationStrategy] = None
    useConvHmx: Optional[bool] = None
    useFoldRelu: Optional[bool] = None
    vtcm_size: Optional[int] = None
    num_hvx_threads: Optional[int] = None


class TFLiteSettings(BaseModel):
    XNNPACK: Optional[XNNPACKSettings] = None
    GPU: Optional[GPUSettings] = None
    QNN: Optional[QNNSettings] = None



"""
Response wrapper models for benchmark and prediction operations.
"""

from typing import List, Union
from pydantic import BaseModel

from .benchmark_result import BenchmarkResult
from .prediction import PredictionResult


class BenchmarkResponse(BaseModel):
    """
    Response from benchmark operation including results and job metadata.
    """
    
    # The actual benchmark results
    results: Union[BenchmarkResult, List[BenchmarkResult]]
    
    # Job tracking metadata
    all_job_ids: List[str] = []
    completed_job_ids: List[str] = []
    incomplete_job_ids: List[str] = []
    
    @property
    def has_incomplete_jobs(self) -> bool:
        """Check if there are incomplete jobs."""
        return len(self.incomplete_job_ids) > 0
    
    @property
    def total_jobs(self) -> int:
        """Total number of jobs submitted."""
        return len(self.all_job_ids)
    
    @property
    def completed_count(self) -> int:
        """Number of completed jobs."""
        return len(self.completed_job_ids)
    
    @property
    def incomplete_count(self) -> int:
        """Number of incomplete jobs."""
        return len(self.incomplete_job_ids)
    
    @property
    def completion_rate(self) -> float:
        """Completion rate as a percentage (0-100)."""
        if self.total_jobs == 0:
            return 100.0
        return (self.completed_count / self.total_jobs) * 100.0


class PredictionResponse(BaseModel):
    """
    Response from prediction operation including results and job metadata.
    """
    
    # The actual prediction results
    results: Union[PredictionResult, List[PredictionResult]]
    
    # Job tracking metadata
    all_job_ids: List[str] = []
    completed_job_ids: List[str] = []
    incomplete_job_ids: List[str] = []
    
    @property
    def has_incomplete_jobs(self) -> bool:
        """Check if there are incomplete jobs."""
        return len(self.incomplete_job_ids) > 0
    
    @property
    def total_jobs(self) -> int:
        """Total number of jobs submitted."""
        return len(self.all_job_ids)
    
    @property
    def completed_count(self) -> int:
        """Number of completed jobs."""
        return len(self.completed_job_ids)
    
    @property
    def incomplete_count(self) -> int:
        """Number of incomplete jobs."""
        return len(self.incomplete_job_ids)
    
    @property
    def completion_rate(self) -> float:
        """Completion rate as a percentage (0-100)."""
        if self.total_jobs == 0:
            return 100.0
        return (self.completed_count / self.total_jobs) * 100.0


"""
Benchmark-related models.
"""

from decimal import Decimal
from typing import Any, Dict, List, Optional

from pydantic import BaseModel

from .device import Device
from .job import BenchmarkStatus
from .settings import Framework


class BenchmarkData(BaseModel):
    """Data from a single benchmark run."""

    Success: Optional[bool] = None
    Status: Optional[BenchmarkStatus] = None
    ComputeUnit: str

    # load
    LoadMsArray: Optional[List[Decimal]] = None
    LoadMsAverage: Optional[Decimal] = None
    LoadMsMedian: Optional[Decimal] = None
    # total inference
    InferenceMsArray: Optional[List[Decimal]] = None
    InferenceMsAverage: Optional[Decimal] = None
    InferenceMsMedian: Optional[Decimal] = None
    # GenAI inference
    PrefillTokens: Optional[int] = None
    GenerationTokens: Optional[int] = None
    PrefillTPS: Optional[Decimal] = None
    GenerateTPS: Optional[Decimal] = None
    # peak ram usage
    PeakLoadRamUsage: Optional[Decimal] = None
    # TODO: inference, named "PeakRamUsage" for legacy support
    PeakRamUsage: Optional[Decimal] = None
    # peak genai ram usage
    PeakPrefillRamUsage: Optional[Decimal] = None
    PeakGenerateRamUsage: Optional[Decimal] = None

    FailureReason: Optional[str] = None
    FailureError: Optional[str] = None
    Stdout: Optional[str] = None
    Stderr: Optional[str] = None

    OutputTensorsId: Optional[str] = None

    Versions: Optional[Dict[str, str]] = None
    Settings: Optional[Dict[str, Any]] = None

    def to_json_dict(self) -> Dict:
        """
        Convert to JSON-friendly dictionary.
        Needed for post requests where Decimals need to be strings.
        """
        from ..utils.json import decimal_list_to_str, decimal_to_str

        result = {
            "Success": self.Success,
            "Status": self.Status,
            "FailureReason": self.FailureReason,
            "FailureError": self.FailureError,
            "Stdout": self.Stdout,
            "Stderr": self.Stderr,
            "ComputeUnit": self.ComputeUnit,
            "LoadMsArray": decimal_list_to_str(self.LoadMsArray),
            "LoadMsAverage": decimal_to_str(self.LoadMsAverage),
            "LoadMsMedian": decimal_to_str(self.LoadMsMedian),
            "InferenceMsArray": decimal_list_to_str(self.InferenceMsArray),
            "InferenceMsAverage": decimal_to_str(self.InferenceMsAverage),
            "InferenceMsMedian": decimal_to_str(self.InferenceMsMedian),
            "PrefillTokens": self.PrefillTokens,
            "GenerationTokens": self.GenerationTokens,
            "PrefillTPS": decimal_to_str(self.PrefillTPS),
            "GenerateTPS": decimal_to_str(self.GenerateTPS),
            "PeakLoadRamUsage": decimal_to_str(self.PeakLoadRamUsage),
            "PeakRamUsage": decimal_to_str(self.PeakRamUsage),
            "PeakPrefillRamUsage": decimal_to_str(self.PeakPrefillRamUsage),
            "PeakGenerateRamUsage": decimal_to_str(self.PeakGenerateRamUsage),
            "Versions": self.Versions,
            "Settings": self.Settings,
        }

        # Remove None values
        return {k: v for k, v in result.items() if v is not None}

    def to_float_dict(self) -> Dict:
        """
        Convert to dictionary with Decimal values converted to float.
        Useful for easier data manipulation and display.
        """

        def decimal_to_float(value):
            if value is None:
                return None
            if isinstance(value, Decimal):
                return float(value)
            return value

        def decimal_list_to_float(value_list):
            if value_list is None:
                return None
            return [float(v) if isinstance(v, Decimal) else v for v in value_list]

        return {
            "Success": self.Success,
            "Status": self.Status,
            "FailureReason": self.FailureReason,
            "FailureError": self.FailureError,
            "Stdout": self.Stdout,
            "Stderr": self.Stderr,
            "ComputeUnit": self.ComputeUnit,
            "LoadMsArray": decimal_list_to_float(self.LoadMsArray),
            "LoadMsAverage": decimal_to_float(self.LoadMsAverage),
            "LoadMsMedian": decimal_to_float(self.LoadMsMedian),
            "InferenceMsArray": decimal_list_to_float(self.InferenceMsArray),
            "InferenceMsAverage": decimal_to_float(self.InferenceMsAverage),
            "InferenceMsMedian": decimal_to_float(self.InferenceMsMedian),
            "PrefillTokens": self.PrefillTokens,  # int, no conversion needed
            "GenerationTokens": self.GenerationTokens,  # int, no conversion needed
            "PrefillTPS": decimal_to_float(self.PrefillTPS),
            "GenerateTPS": decimal_to_float(self.GenerateTPS),
            "PeakLoadRamUsage": decimal_to_float(self.PeakLoadRamUsage),
            "PeakInferenceRamUsage": decimal_to_float(self.PeakRamUsage),
            "PeakPrefillRamUsage": decimal_to_float(self.PeakPrefillRamUsage),
            "PeakGenerateRamUsage": decimal_to_float(self.PeakGenerateRamUsage),
            "OutputTensorsId": self.OutputTensorsId,
            "Versions": self.Versions,
            "Settings": self.Settings,
        }


class BenchmarkDataFloat(BaseModel):
    """Benchmark data with float values instead of Decimal for easier use."""

    Success: Optional[bool] = None
    Status: Optional[BenchmarkStatus] = None
    ComputeUnit: str

    # load
    LoadMsArray: Optional[List[float]] = None
    LoadMsAverage: Optional[float] = None
    LoadMsMedian: Optional[float] = None
    # traditional inference
    InferenceMsArray: Optional[List[float]] = None
    InferenceMsAverage: Optional[float] = None
    InferenceMsMedian: Optional[float] = None
    # GenAI inference
    PrefillTokens: Optional[int] = None
    GenerationTokens: Optional[int] = None
    PrefillTPS: Optional[float] = None
    GenerateTPS: Optional[float] = None
    # peak ram usage (converted to float)
    PeakLoadRamUsage: Optional[float] = None
    PeakInferenceRamUsage: Optional[float] = None
    PeakPrefillRamUsage: Optional[float] = None
    PeakGenerateRamUsage: Optional[float] = None

    FailureReason: Optional[str] = None
    FailureError: Optional[str] = None
    Stdout: Optional[str] = None
    Stderr: Optional[str] = None

    OutputTensorsId: Optional[str] = None

    Versions: Optional[Dict[str, str]] = None
    Settings: Optional[Dict[str, Any]] = None

    @classmethod
    def from_benchmark_data(cls, bd: BenchmarkData) -> "BenchmarkDataFloat":
        """Create a BenchmarkDataFloat from a BenchmarkData instance."""
        return cls(**bd.to_float_dict())


class BenchmarkDbItem(BaseModel):
    """Benchmark database item."""

    UploadId: str
    DeviceInfo: Optional[Device] = None
    Status: BenchmarkStatus
    BenchmarkData: List[BenchmarkData]
    RuntimeFramework: Optional[Framework] = None

    def to_dict(self):
        """Convert to dictionary."""
        d = self.model_dump()
        return d



from enum import Enum
from typing import Optional

from pydantic import BaseModel

from .common import NumThreads


class GraphOptimizationLevel(str, Enum):
    NONE = "none"
    BASIC = "basic"
    EXTENDED = "extended"
    ALL = "all"


class ExecutionMode(str, Enum):
    SEQUENTIAL = "sequential"
    PARALLEL = "parallel"


class CoreMLModelFormat(str, Enum):
    NEURAL_NETWORK = "NeuralNetwork"
    ML_PROGRAM = "MLProgram"


class CoreMLSpecializationStrategy(str, Enum):
    DEFAULT = "Default"
    FAST_PREDICTION = "FastPrediction"


class QNNHtpPerformanceMode(str, Enum):
    BURST = "burst"
    BALANCED = "balanced"
    HIGH_PERFORMANCE = "high_performance"
    HIGH_POWER_SAVER = "high_power_saver"
    LOW_BALANCED = "low_balanced"
    LOW_POWER_SAVER = "low_power_saver"
    SUSTAINED_HIGH_PERFORMANCE = "sustained_high_performance"


class XNNPACKEpSettings(BaseModel):
    intra_op_num_threads: Optional[int | NumThreads] = None


class OpenVINOEpSettings(BaseModel):
    num_of_threads: Optional[int | NumThreads] = None
    enable_qdq_optimizer: Optional[bool] = None
    disable_dynamic_shapes: Optional[bool] = None


class CoreMLEpSettings(BaseModel):
    ModelFormat: Optional["CoreMLModelFormat"] = None
    RequireStaticInputShapes: Optional[bool] = None
    EnableOnSubgraphs: Optional[bool] = None
    SpecializationStrategy: Optional[CoreMLSpecializationStrategy] = None
    AllowLowPrecisionAccumulationOnGPU: Optional[bool] = None


class QNNEpSettings(BaseModel):
    htp_performance_mode: Optional[QNNHtpPerformanceMode] = None
    htp_graph_finalization_optimization_mode: Optional[int] = None
    enable_htp_fp16_precision: Optional[bool] = None
    offload_graph_io_quantization: Optional[bool] = None


class OnnxSettings(BaseModel):
    intra_op_num_threads: Optional[int | NumThreads] = None
    inter_op_num_threads: Optional[int | NumThreads] = None
    allow_intra_op_spinning: Optional[bool] = None
    allow_inter_op_spinning: Optional[bool] = None
    disable_mem_pattern: Optional[bool] = None
    graph_optimization_level: Optional[GraphOptimizationLevel] = None
    execution_mode: Optional[ExecutionMode] = None

    # Execution Provider specific settings
    XNNPACK: Optional[XNNPACKEpSettings] = None
    OpenVINO: Optional[OpenVINOEpSettings] = None
    CoreML: Optional[CoreMLEpSettings] = None
    QNN: Optional[QNNEpSettings] = None

    def format(self):
        return {"Onnx": self.model_dump(exclude_unset=True)}



from .benchmark import (
    BenchmarkData,
    BenchmarkDataFloat,
    BenchmarkDbItem,
)
from .benchmark_result import BenchmarkResult
from .device import Device, DeviceUsage
from .job import JobResult, JobType, BenchmarkStatus
from .model import LicenseInfo, UploadDbItem, UploadedModelType
from .prediction import PredictionResult
from .responses import BenchmarkResponse, PredictionResponse
from .settings import Framework, RuntimeSettings, BenchmarkRequest
from .tensor import IOTensorsMetadata, IOTensorsPresignedUrlResponse, IOType, TensorInfo

__all__ = [
    "Device",
    "DeviceUsage",
    "BenchmarkStatus",
    "BenchmarkData",
    "BenchmarkDataFloat",
    "BenchmarkDbItem",
    "BenchmarkResult",
    "BenchmarkResponse",
    "RuntimeSettings",
    "BenchmarkRequest",
    "Framework",
    "IOType",
    "TensorInfo",
    "IOTensorsMetadata",
    "IOTensorsPresignedUrlResponse",
    "JobResult",
    "JobType",
    "PredictionResult",
    "PredictionResponse",
    "UploadDbItem",
    "UploadedModelType",
    "LicenseInfo",
]



"""
Device selection and filtering logic.
"""

import random
from typing import List, Optional, Union

from rich.console import Console
from rich.table import Table

from runlocal_hub.models.benchmark import Framework

from ..exceptions import DeviceNotAvailableError
from ..http import HTTPClient
from ..models import DeviceUsage
from ..utils.decorators import handle_api_errors
from .filters import DeviceFilters


class DeviceSelector:
    """
    Handles device selection and filtering operations.
    """

    def __init__(self, http_client: HTTPClient):
        """
        Initialize the device selector.

        Args:
            http_client: HTTP client for API requests
        """
        self.http_client = http_client

    @handle_api_errors
    def list_all_devices(
        self, model_id: Optional[str] = None, framework: Optional[Framework] = None
    ) -> List[DeviceUsage]:
        """
        Get a list of available devices for benchmarking.

        Args:
            model_id: Optional ID of a model to get compatible devices
            framework: Optional manual override of runtime framework

        Returns:
            List of available devices with their compute units

        Raises:
            ModelNotFoundError: If the model ID is not found
        """
        endpoint = "/devices/benchmark"
        if model_id is not None:
            endpoint += f"?upload_id={model_id}"

        if framework is not None:
            endpoint += f"&framework={framework.value}"

        response = self.http_client.get(endpoint)

        # Filter to only include devices with compute units if model_id is provided
        if model_id:
            return [
                DeviceUsage(
                    device=device_usage["device"],
                    compute_units=device_usage["compute_units"],
                    native_device_id=device_usage["native_device_id"],
                )
                for device_usage in response
                if device_usage.get("device")
                and device_usage.get("device").get("Disabled") is False
                and device_usage.get("compute_units")
                and len(device_usage.get("compute_units", [])) > 0
            ]

        # Return all devices if no model_id
        return [
            DeviceUsage(
                device=device_usage["device"],
                compute_units=[],
                native_device_id=device_usage["native_device_id"],
            )
            for device_usage in response
            if device_usage.get("device")
            and device_usage.get("device").get("Disabled") is False
        ]

    def select_devices(
        self,
        model_id: str,
        framework: Optional[Framework] = None,
        filters: Optional[Union[DeviceFilters, List[DeviceFilters]]] = None,
        count: Optional[int] = 1,
        user_models: Optional[List[str]] = None,
    ) -> List[DeviceUsage]:
        """
        Select devices based on filter criteria.

        Args:
            model_id: ID of the model to get compatible devices for
            filters: Device filtering criteria. Can be:
                - None: No filtering (all available devices)
                - DeviceFilters: Single filter set
                - List[DeviceFilters]: Multiple filter sets with OR logic (union)
            count: Number of devices to select (default: 1, None = all matching devices)
            user_models: Optional list of user's models for validation

        Returns:
            List of matching devices

        Raises:
            ValueError: If model_id doesn't belong to user
            DeviceNotAvailableError: If no devices match the criteria
        """
        # Validate model_id if user_models provided
        if user_models is not None and model_id not in user_models:
            from ..exceptions import ModelNotFoundError

            available_models_str = f"Available models: {', '.join(user_models[:5])}"
            if len(user_models) > 5:
                available_models_str += f" ... and {len(user_models) - 5} more"

            raise ModelNotFoundError(
                f"Model '{model_id}' not found in your account. {available_models_str}",
                model_id=model_id,
                available_models=user_models,
            )

        # Get all available devices for this model
        all_devices = self.list_all_devices(model_id=model_id, framework=framework)

        # Handle different filter types
        if filters is None:
            # No filtering - use all devices
            filtered_devices = all_devices
        elif isinstance(filters, DeviceFilters):
            # Single filter set
            filtered_devices = self._apply_filters(all_devices, filters)

            # Check if any devices matched
            if not filtered_devices:
                # Create helpful error message with filter details
                filter_details = self._get_filter_details(filters)
                filter_description = ", ".join(
                    [f"{k}={v}" for k, v in filter_details.items()]
                )

                error_message = (
                    f"No devices match the specified criteria ({filter_description}). "
                )
                error_message += (
                    f"Found {len(all_devices)} total devices for this model. "
                )
                error_message += "Try relaxing your filter conditions."

                raise DeviceNotAvailableError(
                    error_message,
                    filters_used=filter_details,
                    available_count=len(all_devices),
                )
        else:
            # Multiple filter sets with OR logic
            seen_device_ids = set()
            filtered_devices = []

            for filter_set in filters:
                # Apply this filter set
                matches = self._apply_filters(all_devices, filter_set)

                # Add unique devices to result
                for device_usage in matches:
                    if device_usage.native_device_id not in seen_device_ids:
                        seen_device_ids.add(device_usage.native_device_id)
                        filtered_devices.append(device_usage)

            # Check if any devices matched
            if not filtered_devices:
                # Create helpful error message with all filter details
                all_filter_details = []
                for i, filter_set in enumerate(filters):
                    filter_details = self._get_filter_details(filter_set)
                    if filter_details:
                        filter_description = ", ".join(
                            [f"{k}={v}" for k, v in filter_details.items()]
                        )
                        all_filter_details.append(
                            f"Filter {i + 1}: {filter_description}"
                        )

                error_message = (
                    "No devices match any of the specified criteria:\n"
                    + "\n".join(all_filter_details)
                    + "\n"
                )
                error_message += (
                    f"Found {len(all_devices)} total devices for this model. "
                )
                error_message += "Try relaxing your filter conditions."

                raise DeviceNotAvailableError(
                    error_message,
                    filters_used={"filters": all_filter_details},
                    available_count=len(all_devices),
                )

        # Apply count logic: None means all devices, otherwise limit to count
        if count is not None and len(filtered_devices) > count:
            # Randomly select 'count' devices from the filtered list
            filtered_devices = random.sample(filtered_devices, count)

        return filtered_devices

    def _get_filter_details(self, filters: DeviceFilters) -> dict:
        """
        Extract filter details for error messages.

        Args:
            filters: Device filters to extract details from

        Returns:
            Dictionary of filter details
        """
        filter_details = {}
        if filters.device_name:
            filter_details["device_name"] = filters.device_name
        if filters.soc:
            filter_details["soc"] = filters.soc
        if filters.ram_min:
            filter_details["ram_min"] = f"{filters.ram_min}GB"
        if filters.ram_max:
            filter_details["ram_max"] = f"{filters.ram_max}GB"
        if filters.year_min:
            filter_details["year_min"] = filters.year_min
        if filters.year_max:
            filter_details["year_max"] = filters.year_max
        if filters.os:
            filter_details["os"] = filters.os
        if filters.compute_units:
            filter_details["compute_units"] = filters.compute_units
        return filter_details

    def _apply_filters(
        self,
        devices: List[DeviceUsage],
        filters: DeviceFilters,
    ) -> List[DeviceUsage]:
        """
        Apply filter criteria to a list of devices.

        Args:
            devices: List of devices to filter
            filters: Filter criteria

        Returns:
            Filtered list of devices
        """
        filtered = devices

        # Filter by device name (substring match)
        if filters.device_name is not None:
            filtered = [
                d
                for d in filtered
                if filters.device_name.lower() in d.device.Name.lower()
            ]

        # Filter by SoC (substring match)
        if filters.soc is not None:
            filtered = [
                d for d in filtered if filters.soc.lower() in d.device.Soc.lower()
            ]

        # Filter by RAM range
        if filters.ram_min is not None:
            filtered = [d for d in filtered if d.device.Ram >= filters.ram_min]

        if filters.ram_max is not None:
            filtered = [d for d in filtered if d.device.Ram <= filters.ram_max]

        # Filter by year range
        if filters.year_min is not None:
            filtered = [d for d in filtered if d.device.Year >= filters.year_min]

        if filters.year_max is not None:
            filtered = [d for d in filtered if d.device.Year <= filters.year_max]

        # Filter by OS (substring match)
        if filters.os is not None:
            filtered = [
                d for d in filtered if filters.os.lower() in d.device.OS.lower()
            ]

        # Filter by compute units - only keep specified compute units
        if filters.compute_units is not None:
            new_filtered = []
            for device_usage in filtered:
                # Keep only compute units that are in the filter
                matching_compute_units = [
                    cu
                    for cu in device_usage.compute_units
                    if cu in filters.compute_units
                ]

                # Only include device if it has at least one matching compute unit
                if matching_compute_units:
                    new_filtered.append(
                        DeviceUsage(
                            device=device_usage.device,
                            compute_units=matching_compute_units,
                            native_device_id=device_usage.native_device_id,
                        )
                    )

            filtered = new_filtered

        return filtered

    def display_selected_devices(self, devices: List[DeviceUsage]) -> None:
        """
        Display selected devices in a nice format using rich.

        Args:
            devices: List of selected devices to display
        """
        console = Console()

        table = Table(
            title=f"Selected [cyan]{len(devices)}[/cyan] device(s)",
            title_style="bold",
            show_header=True,
            header_style="bold magenta",
            expand=True,
        )
        table.add_column("Device")
        table.add_column("Year", justify="center")
        table.add_column("SoC", style="cyan")
        table.add_column("RAM", style="dim", justify="right")
        table.add_column("OS", style="dim")
        table.add_column("Compute Units", style="green")

        for device in devices:
            table.add_row(
                device.device.Name,
                str(device.device.Year),
                device.device.Soc,
                f"{device.device.Ram}GB",
                f"{device.device.OS} {device.device.OSVersion}",
                ", ".join(device.compute_units),
            )

        console.print(table)

        print("")
